#include <sys/stat.h>
#include <termios.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <fcntl.h>
#include <inttypes.h>
#include <string>
#include <errno.h>
#include "node-comm.hh"
#include "crc.h"

/**************************************************************************
 *
 * struct timeval functions
 *
 *************************************************************************/
const int usecs_pr_sec = 1000000;

void normalize_timeval(struct timeval &res)
{
  while (res.tv_usec > usecs_pr_sec) {
    res.tv_usec -= usecs_pr_sec;
    res.tv_sec += 1;
  }

  while (res.tv_usec < 0) {
    res.tv_usec += usecs_pr_sec;
    res.tv_sec -= 1;
  }
}

struct timeval operator+(const struct timeval& a, const struct timeval &b)
{
  struct timeval res;
  
  res.tv_sec = a.tv_sec + b.tv_sec;
  res.tv_usec = a.tv_usec + b.tv_usec;

  normalize_timeval(res);
  
  return res;
}

struct timeval operator-(const struct timeval& a, 
			 const struct timeval& b)
{
  struct timeval res;
  
  res.tv_sec = a.tv_sec - b.tv_sec;
  res.tv_usec = a.tv_usec - b.tv_usec;

  normalize_timeval(res);
  
  return res;
}

double tv2f(const struct timeval &t)
{
  return (double)t.tv_sec + (double)t.tv_usec / (double)usecs_pr_sec;
}

/**************************************************************************
 *
 * Input file handling
 *
 *************************************************************************/

bool parse_line(char *buffer, int *res)
{
  char *colon = buffer;

  for (int i = 0; i < 6; i++) {
    colon = strchr(colon, ';');
    
    if (!colon)
      return false;

    colon++;

    res[i] = atoi(colon);
  }

  return true;
}

int count_lines = 0;

bool fill_buffer(FILE *f, uint16_t *buf)
{
  char buffer[1024];

  uint16_t *buf_pos = buf;

  for (int i = 0; i < 42; ) {
    if (fgets(buffer, 1024, f)) {
      int val[6];
      
      if (parse_line(buffer, val)) {
	*buf_pos++ = val[1];
	*buf_pos++ = val[2];
	*buf_pos++ = val[3];
	i++;
	count_lines++;
      }
    } else {
      return false;
    }
  }
  return true;
}

/**************************************************************************
 *
 * Serial communication
 *
 *************************************************************************/
char line[1024];
char *linepos = line;

void flush_serial(int fd)
{
  linepos = line;
  read(fd, line, 1024);
}

char *read_line(int fd, struct timeval *tv)
{
  static char res[1024];
  fd_set rfds;

  /* Do we already have a new line? */
  char * pos;
  
  pos = strchr(line, '\n');
  if (pos) {
    *pos = '\0';
    
    strncpy(res, line, 1024);
    
    memmove(line, pos + 1, 1024 - (pos - line) - 1);
    linepos -= pos - line + 1;
    return res;
  }

  while (1) {
    FD_ZERO(&rfds);
    FD_SET(fd, &rfds);
    int retval = select(fd + 1, &rfds, NULL, NULL, tv);
    if (retval != 0) {
      retval = read(fd, linepos, 1024 - (linepos - line));
      if (retval) {
	linepos += retval;
	*linepos = '\0';
	
	pos = strchr(line, '\n');
	if (pos) {
	  *pos = '\0';
	  
	  strncpy(res, line, 1024);
	  
	  memmove(line, pos + 1, 1024 - (pos - line) - 1);
	  linepos -= pos - line + 1;
	  return res;
	}
      }
    } else {
      return 0;
    }
  } 
}

uint8_t *read_256bytes(int fd)
{
  static uint8_t buffer[256];
  uint8_t *buffer_pos = buffer;
  int retval;
  fd_set rfds;

  while (buffer_pos != buffer + 256) {
    FD_ZERO(&rfds);
    FD_SET(fd, &rfds);
    retval = select(fd + 1, &rfds, NULL, NULL, 0);
    if (retval != 0) {
      retval = read(fd, buffer_pos, 256 - (buffer_pos - buffer));
      
      if (retval) {
	buffer_pos += retval;
      }
    }
  }
  return buffer;
}

uint8_t *read_block(int fd)
{
  const int block_size = 259;
  static uint8_t buffer[block_size];
  uint8_t *buffer_pos = buffer;
  int retval;
  fd_set rfds;
  struct timeval tv;

  memset(buffer, 0xAA, block_size);

  while (buffer_pos - buffer < block_size) {
    FD_ZERO(&rfds);
    FD_SET(fd, &rfds);
    if (buffer_pos == buffer) {
      tv.tv_sec = 1;
      tv.tv_usec = 0;
    } else {
      tv.tv_sec = 5;
      tv.tv_usec = 0;
    }

    retval = select(fd + 1, &rfds, NULL, NULL, &tv);
    if (retval > 0) {
      retval = read(fd, buffer_pos, block_size - (buffer_pos - buffer));
      
      if (retval > 0) {
	buffer_pos += retval;
	if ((buffer_pos - buffer) == 2 && buffer[0] == 'f' 
	    && buffer[1] == '\n') {
	  return 0;
	} 
      } else if (retval < 0) {
	throw std::string(std::string("Error reading: ") + strerror(errno));
      } else if (retval == 0) {
	fprintf(stderr, "Strange, read 0 bytes?\n");
      }
    } else if (retval == 0) {
      if (buffer_pos == buffer) {
	write(fd, "!", 1);
      } else {
	throw std::string("Timeout reading block");
      }
    } else {
      throw std::string("Select error");
    }
  }

  uint16_t crc = 0xFFFF;
  
  for (size_t j = 0; j < block_size - 2; j++)
    crc = crcByte(crc, ((uint8_t*)buffer)[j]);

  if (crc != *(uint16_t*)(buffer + block_size - 2)) {
    char buf[1024];
    snprintf(buf, 1024, "CRC in packet is 0x%04X. Should have been 0x%04X\n", 
	     *(uint16_t*)(buffer + block_size - 2), crc);
    throw std::string(buf);
  }

  return buffer + 1;
}

uint16_t real_buffer[4 * 239 * 256 / sizeof(uint16_t)];

void sync_with_node(int fd)
{
  printf("Syncing with node... ");
  fflush(stdout);

  while (1) {
    flush_serial(fd);
    write(fd, "p", 1);
    struct timeval tv;
    tv.tv_sec = 1;
    tv.tv_usec = 0;

    char *line = read_line(fd, &tv);
    if (line == 0) {
      uint8_t buf[256];
      memset(buf, 0, 256);
      write(fd, buf, 256);
      continue;
    }
      
    if (strcmp(line, "pong") == 0) {
      break;
    }
  }

  printf("done\n");
}

void flush_node(int fd) 
{
  printf("Flushing... ");
  write(fd, "f", 2);
  char *line = read_line(fd, 0);

  if (strcmp(line, "done") != 0) {
    printf("Got \"%s\" instead of done. Bailing out\n", line);
    exit(1);
  }
  printf("done\n");  
}

const uint16_t total_pages = (0xFFFF - 0x1100) / 256 * 4;

void really_write(int fd, const void *buf, size_t len)
{
  size_t left = len;
  do {
    size_t res = write(fd, (const uint8_t*)buf + (len - left), left);
    if (res > 0) {
      left -= res;
    } else if (res < 0) {
      if (errno == EINTR) {
	continue;
      } else {
	perror("Error writing to node");
	    exit(1);
      }
    }
  } while (left > 0);
}

int uploaded_pages;

bool upload_pages(int fd, FILE *data_file)
{
  struct timeval tv;
  tv.tv_sec = 1;
  tv.tv_usec = 0;
  // Read potential left-over data from the node.
  char *line = read_line(fd, &tv);

  printf("Preparing for data upload... ");
  write(fd, "a", 2);
  line = read_line(fd, 0);

  if (strcmp(line, "ok") != 0) {
    printf("Got \"%s\" instead of ok. Bailing out\n", line);
    exit(1);
  }

  printf("done\n");

  printf("Uploading pages");

  uint16_t *buffer = real_buffer;

  bool more_data = true;

  uploaded_pages = 0;

  for (int i = 0; i < total_pages; i++) {
    bool retry = false;

    more_data = fill_buffer(data_file, buffer);
    do {
      if (retry) {
	sleep(1);
	flush_serial(fd);
      }
      really_write(fd, "b", 1);
      really_write(fd, buffer, 256);
	    
      uint16_t crc = 0xFFFF;

      for (size_t j = 0; j < 256; j++)
	crc = crcByte(crc, ((uint8_t*)buffer)[j]);

      really_write(fd, &crc, 2);

      struct timeval tv;
      tv.tv_sec = 1;
      tv.tv_usec = 0;
      
      char * line = read_line(fd, &tv);
      if (!line) {
	printf("Node did not respond in a timely manner. Trying again.\n");
	char buf[256];
	memset(buf, 0, 256);
	write(fd, buf, 256);
	read_line(fd, &tv);
	retry = true;
      } else {
	if (strcmp(line, "ok") != 0) {
	  retry = true;
	  printf("Node responded with \"%s\". Trying again.\n", line);
	} else {
	  printf(".");
	  fflush(stdout);
	  retry = false;
	  uploaded_pages++;
	}
      }
    } while (retry);
    buffer += 256 / sizeof(*buffer);
    if (!more_data)
      break;
  }

  write(fd, "x", 1);
  printf(" done\n");
  return more_data;
}

void read_uploaded_pages(int fd)
{
  printf("Reading pages back");

  uint16_t *buffer = real_buffer;
  
  for (int i = 0; i < uploaded_pages; i++) {
    write(fd, "r", 1);
    
    uint8_t *buf = read_256bytes(fd);
    
    if (memcmp(buffer, buf, 256) == 0) {
      printf(".");
      fflush(stdout);
    } else {
      printf("\nPosition %d did not match\n", i);

      printf("Data sent to node:\n");

      for (int j = 0; j < 256; j++) {
	printf("%02x ", ((uint8_t*)buffer)[j]);
	if ((j + 1) % 16 == 0)
	  printf("\n");
      }

      printf("\n");

      printf("Data received from node:\n");

      for (int j = 0; j < 256; j++) {
	printf("%02x ", ((uint8_t*)buf)[j]);
	if ((j + 1) % 16 == 0)
	  printf("\n");
      }

      printf("\n");
    }
    buffer += 256 / sizeof(*buffer);
  }
  
  write(fd, "x", 1);
  
  printf("\n");
}

double total_time = 0;
double total_watt = 0;
int count_watt = 0;

void do_compress(int fd, FILE *out_file, bool cont)
{
  struct timeval start_time, end_time;

  if (out_file) {
    write(fd, "C", 1);
  } else {
    write(fd, "c", 1);
  }
  start_power();
  gettimeofday(&start_time, 0);

  printf("Compressing data");
  fflush(stdout);
  
  uint8_t *data = 0;
  do {
    try {
      data = read_block(fd);
    } catch (std::string &e) {
      printf(("\nLost sync with node (error: " + e 
	      + "). Flushing, and requesting resend.\n").c_str());
      sleep(1);
      uint8_t buf[1024];
      read(fd, buf, 1024);
      write(fd, "a", 1);
      data = buf;
      continue;
    }
    if (out_file && data) {
      write(fd, "o", 1);
      fwrite(data, 256, 1, out_file);
      printf(".");
      fflush(stdout);
    }
  } while (data != 0);

  gettimeofday(&end_time, 0);
  double power = stop_power();

  printf("\nReceived end signal\n");
  fflush(stdout);

  printf("Compression took: %6.4f seconds. Average power: %6.4f mW\n", 
	 tv2f(end_time - start_time), power);

  total_time += tv2f(end_time - start_time);
  total_watt += power;
  count_watt++;

  if (!cont && out_file) {
    write(fd, "F", 1);
    try {
      data = read_block(fd);
      fflush(stdout);
    } catch (std::string &e) {
      printf(("Error: " + e + "\n").c_str());
      exit(0);
    }
    if (data) {
      fwrite(data, 256, 1, out_file);
    }
  }
}

int main(int argc, char * argv[]) 
{
  struct termios oldtio, newtio;
  int fd;
  FILE *data_file;
  FILE *out_file = 0;

  if (argc < 2) {
    fprintf(stderr, "You must specify a file to upload\n");
    exit(1);
  }

  if (argc == 3) {
    out_file = fopen(argv[2], "w");
    if (!out_file) {
      perror("Cannot open output file");
    }
  }

  data_file = fopen(argv[1], "r");
  if (!data_file) {
    perror("Cannot open data file");
    exit(1);
  }

  fd = open("/dev/ttyUSB0", O_RDWR);
  if (fd < 0) {
    perror("Cannot open serial port");
    exit(1);
  }

  tcgetattr(fd, &oldtio);	/* save current serial port settings */
  memset(&newtio, 0, sizeof (newtio));

  newtio.c_cflag = B230400 | CS8 | CLOCAL | CREAD | HUPCL;
  newtio.c_iflag = IGNBRK;
  newtio.c_oflag = 0;
  newtio.c_lflag = 0;//FLUSHO;

  //  tcflush(fd, TCIOFLUSH);
  tcsetattr(fd, TCSANOW, &newtio);

  tcflush(fd, TCIOFLUSH);

  init_power();

  sync_with_node(fd);

  flush_node(fd);

  bool cont = true;
  while (cont) {
    cont = upload_pages(fd, data_file);
    read_uploaded_pages(fd);
    do_compress(fd, out_file, cont);
  }

  printf("End results:\n");
  printf("  Compressed lines: %d\n", count_lines);
  printf("  Compression time: %6.4f s\n", total_time);
  printf("  Average Power   : %6.4f mW\n", total_watt / count_watt);

  halt_power();

  return 0;
}
